/*
 * oracle C datatypes OCINumber, OCIDate
 * see also the conversion functions 
 * OCIDateFromText
 * OCIDateToText
 *
 *
 * changes:
 *
 * 20.05.1999   now handles all types of sql statements "OCIStmtExecute"
 *              (select, insert, update, alter ... )
 *
 *
 */
 
#include <stdio.h>

#include <oci.h>


typedef struct _tagCOLINFO
{
    void      *pParamData;              // ORACLE desc buffer
    ub2        col_type;                // type ot the column
    text      *col_name;                // name of the column
    ub2        buf_len;                 // the needed char's for that datatype
    char      *pBuf;                    // holding the data
    char      *pWork;                   // alloc temp buffer for sprintf
    OCIDefine *pOraDefine;              // ORACLE "define" buffer 
    sb2        nOraIndicator;           // out var from OCIDefineByPos

} COLINFO;


#define MAX_COL_SUPPORTED      500
#define SRC_LOCATION           __FILE__, psz_function_name__, __LINE__
#define FUNCTION(a)            static char psz_function_name__[] = a;

#define ORA_CHECK_RC(a,b)      ORA_DispError(SRC_LOCATION,a,b);


/*char gszStmt[] = "SELECT id,name,vname,dgeb FROM ssv_emp"; */
//char gszStmt[] = "SELECT name,vname,birthday FROM bbstypes1";
char gszStmt[] = "SELECT name,vname FROM bbstypes1";
//char gszStmt[] = "SELECT name,vname,birthday,salary FROM bbstypes1";
//char gszStmt[] = "select password syspass from sys.user$ where name = 'oracle'";
//char gszStmt[] = "insert into bbstypes1(name) values('newname')";


void    i_OCI_print_handle_size( void )
{
   printf( "int        : %d\n", sizeof( int     ) );
   return;
}


/********************************************************************
 *
 * func: rtrim
 *
 ********************************************************************/

void rtrim( char *pszStr )
{
  int 	nLen   = strlen( pszStr );
  char *pszPos = NULL;
  int   ii     = 0;
  
  /* if the last char is not a SPACE just return */
  if ( *(pszStr + nLen -1) != ' ' )
  {
    return;
  }

  /* find the last position where we have a character */  
  pszPos = ( pszStr + nLen -1 );
  while ( ( *pszPos == ' '   )  &&
          ( pszPos != pszStr )      )
  {
    /* jump back one char */
    pszPos--;  
  }
  
  *(pszPos + 1) = '\0';           

}

ub2 ORA_PrintStmtType( OCIStmt *pOraStmt, OCIError *pOraError )
{
   ub2   nType = 0;

   sword ora_rc;

   //
   // get sql type information (select, insert, update, alter...
   //
   ora_rc = OCIAttrGet( pOraStmt,
                        OCI_HTYPE_STMT, 
                        &nType,
                        0,
                        OCI_ATTR_STMT_TYPE,
                        pOraError );

   switch( nType )
   {
      case OCI_STMT_SELECT:
         printf( "SQL statment type is: %s\n", "SELECT" );
         break;
      case OCI_STMT_UPDATE:
         printf( "SQL statment type is: %s\n", "UPDATE" );
         break;
      case OCI_STMT_DELETE:
         printf( "SQL statment type is: %s\n", "DELETE" );
         break;
      case OCI_STMT_INSERT:
         printf( "SQL statment type is: %s\n", "INSERT" );
         break;
      case OCI_STMT_CREATE:
         printf( "SQL statment type is: %s\n", "CREATE" );
         break;
      case OCI_STMT_DROP:
         printf( "SQL statment type is: %s\n", "DROP" );
         break;
      case OCI_STMT_ALTER:
         printf( "SQL statment type is: %s\n", "ALTER" );
         break;
      case OCI_STMT_BEGIN:
         printf( "SQL statment type is: %s\n", "BEGIN" );
         break;
      case OCI_STMT_DECLARE:
         printf( "SQL statment type is: %s\n", "DECLARE" );
         break;

      default:
         printf( "SQL statment type is: %s\n", "UNKOWN (default in switch)" );
   }

   return(nType);
}


/********************************************************************
 *
 * func: ORA_DispError
 *
 ********************************************************************/

void ORA_DispError( char *pszFile, 
                    char *pszFunc, 
                    int nLineNo, 
                    OCIError *errhp, 
                    sword status     )
{
  text errbuf[512];
  ub4 buflen;
  sb4 errcode;

  char *pszOraStatus     = NULL;

  char *ora_status_tbl[] = {  "OCI_SUCCESS" ,              /* index 0 */
                              "OCI_SUCCESS_WITH_INFO",     /*       1 */
                              "OCI_NEED_DATA",             /*       2 */
                              "OCI_NO_DATA",               /*       3 */
                              "OCI_ERROR",                 /*       4 */
                              "OCI_INVALID_HANDLE",        /*       5 */
                              "OCI_STILL_EXECUTING",       /*       6 */
                              "OCI_CONTINUE",              /*       7 */
                              "OCI_unknown"                /*       8 */
                           }; 

  switch (status)
  {
     case OCI_SUCCESS:
       pszOraStatus = ora_status_tbl[0];
       break;

     case OCI_SUCCESS_WITH_INFO:
       pszOraStatus = ora_status_tbl[1];
       break;

     case OCI_NEED_DATA:
       pszOraStatus = ora_status_tbl[2];
       break;

     case OCI_NO_DATA:
       pszOraStatus = ora_status_tbl[3];
       break;

     case OCI_ERROR:
       pszOraStatus = ora_status_tbl[4];
       break;

     case OCI_INVALID_HANDLE:
       pszOraStatus = ora_status_tbl[5];
       break;

     case OCI_STILL_EXECUTING:
       pszOraStatus = ora_status_tbl[6];
       break;

     case OCI_CONTINUE:
       pszOraStatus = ora_status_tbl[7];
       break;

     default:
       pszOraStatus = ora_status_tbl[8];
       break;
   }

   if ( status != OCI_ERROR && status != OCI_NO_DATA )
   {
//      printf( "status: %s\n", pszOraStatus );
//      printf( "file  : %s     line :   %d\n", pszFile, nLineNo );
   }
   else
   {
      OCIErrorGet ( (dvoid *) errhp, (ub4) 1, (text *) NULL, (sb4*)&errcode,
                    errbuf, (ub4) sizeof(errbuf), (ub4) OCI_HTYPE_ERROR      );

      printf( "file  : %s     line :   %d\n", pszFile, nLineNo );
      printf( "error: %s\n text: %s\n", pszOraStatus, errbuf );
   }

}


/********************************************************************
 *
 * func: ORA_SessionSetUser
 *
 ********************************************************************/

int ORA_SessionSetUser( OCISession *pSession, 
                        OCIError   *pOraError,
                        char       *pszUser, 
                        char       *pszPW      )
{
   FUNCTION("ORA_SessionSetUser")
   int    rc           = -1;
   sword	 ora_rc       =  0;

   // set userID
   //
   ora_rc = OCIAttrSet( pSession, OCI_HTYPE_SESSION,
                        pszUser, (ub4)strlen( pszUser ),
                        OCI_ATTR_USERNAME, pOraError     );
   // set PW
   // 
   ora_rc = OCIAttrSet( pSession, OCI_HTYPE_SESSION,
                        pszPW, (ub4)strlen( pszPW ),
                        OCI_ATTR_PASSWORD, pOraError     );

   // set rc to OK in any cases
   rc = 0;

   return(rc);
}



/********************************************************************
 *
 * func: ORA_ServiceSetSession
 *
 ********************************************************************/

int ORA_ServiceSetSession( OCISvcCtx  *pService,
                           OCISession *pSession,
                           OCIError   *pOraError  )
{
   FUNCTION("ORA_ServiceSetSession")
   int    rc           = -1;
   sword	 ora_rc       =  0;

   ora_rc = OCIAttrSet( pService, OCI_HTYPE_SVCCTX,
                        pSession, (ub4) 0,
                        OCI_ATTR_SESSION,
                        pOraError                  );

   // set to OK in any cases
   rc = 0;

   return(rc);
}



/********************************************************************
 *
 *
 * func: main
 *
 *
 *
 ********************************************************************/

int main( void )
{
   FUNCTION("main")

   char         pszUser[]    = { "system" };
   char         pszPW[]      = { "bbstora8"  };
   char         pszDBImage[] = { "bbst"  };


   COLINFO	    colinfo[ MAX_COL_SUPPORTED ];

   /* internal only for info */
   long         nCols        = 0;
   long         nRows        = 0;
   long         nRecLen      = 0;
   char        *pRec         = NULL;

   int          ii           = 0;  
   int          jj           = 0;
   char        *pszTemp      = NULL;


   /* oracle data typtes */
   OCIEnv       *pOraEnv     = NULL;
   OCIServer    *pOraServer  = NULL;
   OCIError     *pOraError   = NULL;
   OCISession   *pOraSession = NULL;
   OCISvcCtx    *pOraService = NULL;    // service context 

   OCIStmt      *pOraStmt    = NULL;    // statement handle


   char          szOraVersion [512];    // store version info from the DBMS
   
   sword	ora_rc       = 0;
   ub2          nStmtType    = 0;
   ub4          nIters       = 0;       // default for select statements

      
   printf( "start...\n" );

   i_OCI_print_handle_size();

   //
   // | OCI_OBJECT
   //   
   ora_rc = OCIInitialize( (ub4) OCI_THREADED |     // thread safe
                                 OCI_OBJECT,        // object support
                                 NULL,
                                 NULL,   	    // no userdef alloc
                                 NULL,   	    // no userdef realloc
                                 NULL          );   // no userdef free

   // init the ORA environment
   //			      			   
   ora_rc = OCIEnvInit( &pOraEnv,
                        OCI_DEFAULT,
                        (size_t) 0,
                        NULL         );
			

   // get server handle
   //			
   ora_rc = OCIHandleAlloc( pOraEnv,
                            (void*)&pOraServer,
                            (ub4) OCI_HTYPE_SERVER,
                            (size_t) 0,
                            NULL                     );


   // get error handle
   //			
   ora_rc = OCIHandleAlloc( pOraEnv,
                            (void*)&pOraError,
                            (ub4) OCI_HTYPE_ERROR,
                            (size_t) 0,
                            NULL                     );

   // get service handle
   // 
   ora_rc = OCIHandleAlloc( pOraEnv,
                            (void*)&pOraService,
                            (ub4) OCI_HTYPE_SVCCTX,
                            (size_t) 0,
                            NULL                     );

   // get session handle
   // 
   ora_rc = OCIHandleAlloc( pOraEnv,
                            (void*)&pOraSession,
                            (ub4) OCI_HTYPE_SESSION,
                            (size_t) 0,
                            NULL                     );

   // get statement handle
   //
   ora_rc = OCIHandleAlloc( pOraEnv,
                            (void*)&pOraStmt,
                            (ub4) OCI_HTYPE_STMT,
                            (size_t) 0,
                            NULL                     );


   // set data source (image name)
   //
   ora_rc = OCIServerAttach( pOraServer,
                             pOraError,
                             (text*)pszDBImage, 
                             (ub4)strlen( pszDBImage ),
                             OCI_DEFAULT                  );

   ORA_CHECK_RC(pOraError,ora_rc)


   // register datasource in the service handle
   //
   ora_rc = OCIAttrSet( pOraService, OCI_HTYPE_SVCCTX,
                        pOraServer, (ub4)0, 
                        OCI_ATTR_SERVER,
                        pOraError );

   (void)ORA_SessionSetUser( pOraSession, pOraError, pszUser, pszPW );

   // authenticate user, start session
   //
   ora_rc = OCISessionBegin( pOraService,
                             pOraError,
                             pOraSession,
                             OCI_CRED_RDBMS,
                        /*     OCI_CRED_EXT,     */     
                             OCI_DEFAULT      );

//   printf( "status after OCISessionBegin: " );
   ORA_CHECK_RC(pOraError,ora_rc)


   (void)ORA_ServiceSetSession( pOraService, pOraSession, pOraError );

//   printf( "status after ORA_ServiceSetSession: " );
   ORA_CHECK_RC(pOraError,ora_rc)



   //
   // get some version infos
   //
   ora_rc = OCIServerVersion( pOraServer, pOraError,
                              (text*)szOraVersion, 
                              sizeof( szOraVersion ),
                              OCI_HTYPE_SERVER       );
   printf( "ORACLE Server Version: %s\n", szOraVersion );




   /*                          */
   /* execute the SQL stmt     */
   /*                          */

   ora_rc = OCIStmtPrepare( pOraStmt,
                            pOraError,
                            (text*)gszStmt,
                            strlen(gszStmt),
                            OCI_NTV_SYNTAX,
                            OCI_DEFAULT      );

//   printf( "status after OCIStmtPrepare: " );
   ORA_CHECK_RC(pOraError,ora_rc)

   nStmtType = ORA_PrintStmtType( pOraStmt, pOraError );
   if( nStmtType != OCI_STMT_SELECT )
   {
      nIters = 1;
   }

   ora_rc = OCIStmtExecute( pOraService,
                            pOraStmt,
                            pOraError,
                            nIters,               // iteration for non SELECT calls, 0 for select..
                            0,
                            (OCISnapshot *)NULL,
                            (OCISnapshot *)NULL,
                            OCI_DEFAULT             );

//   printf( "status after OCIStmtExecute: " );
   ORA_CHECK_RC(pOraError,ora_rc)


   if ( nIters == 1                     && 
        (ora_rc == OCI_SUCCESS          || 
         ora_rc == OCI_SUCCESS_WITH_INFO )   )
   {
       //
       // commit the trx and do cleanup
       //

       OCITransCommit( pOraService, pOraError, OCI_DEFAULT );
       printf( "TRX commited\n");

       goto CLEANUP;       
   }
   if ( nIters == 1                     && 
        ! (ora_rc == OCI_SUCCESS        || 
           ora_rc == OCI_SUCCESS_WITH_INFO )   )
   {
       //
       // commit the trx and do cleanup
       //

       OCITransRollback( pOraService, pOraError, OCI_DEFAULT );
       printf( "TRX rollback\n");

       goto CLEANUP;       
   }



   /*                          */
   /* get statistic info       */
   /*                          */

   //
   // get column count
   //
   ora_rc = OCIAttrGet( pOraStmt, OCI_HTYPE_STMT,
                        &nCols, NULL,
                        OCI_ATTR_PARAM_COUNT, pOraError );
   

   //
   // get row count !!! always return 0 !!!
   //
   ora_rc = OCIAttrGet( pOraStmt, OCI_HTYPE_STMT,
                        &nRows, NULL,
                        OCI_ATTR_ROW_COUNT, pOraError );


//   printf( "status after OCIAttrGet - OCI_ATTR_ROW_COUNT: " );
   ORA_CHECK_RC(pOraError,ora_rc)


   printf( "number of columns: %ld\n", nCols );
   printf( "number of rows   : %ld\n", nRows );

   if ( nCols > MAX_COL_SUPPORTED )
   {
      printf( "panic, max Cols supported = %ld\n", MAX_COL_SUPPORTED );
      exit(1);
   }


   /*                           */
   /* allocate some storage     */
   /*                           */

   // allocate the bind buffers

   for( ii=0; ii < nCols; ii++ )
   {
      ub4   nLen  = 0;
      ub1   nType = 0;
      ub4   nSize = 0;

      // get infos about the column

      ora_rc = OCIParamGet( pOraStmt, OCI_HTYPE_STMT, pOraError,
                            &(colinfo[ii].pParamData), ii+1      );

      // store col type
      ora_rc = OCIAttrGet( colinfo[ii].pParamData, OCI_DTYPE_PARAM, 
                           &(colinfo[ii].col_type), 0,
                           OCI_ATTR_DATA_TYPE, pOraError );

      // get required memory
      ora_rc = OCIAttrGet( colinfo[ii].pParamData, OCI_DTYPE_PARAM, 
                           &(colinfo[ii].buf_len), 0,
                           OCI_ATTR_DATA_SIZE, pOraError );


      // get col name
      // !!! the returned string was not always nul terminated !!!
      // !!! but the "nSize" value was always correct          !!!

      ora_rc = OCIAttrGet( colinfo[ii].pParamData, OCI_DTYPE_PARAM, 
                           (dvoid*)&pszTemp, 
                           &nSize,
                           OCI_ATTR_NAME, pOraError );

      //
      // store the column name in the structure
      //

      // alloc memory for the colname
      colinfo[ii].col_name = (text*)malloc( nSize + 1 );
      if ( colinfo[ii].col_name != NULL ) 
      {
         // OK, copy the colname
         memcpy( colinfo[ii].col_name, pszTemp, nSize );
         // nul term the string
         *(colinfo[ii].col_name + nSize ) = 0;
      }
      else
      {
         printf( "panic, no memory left\n" );
         exit(1);
      }


      //
      // get memory to store the column data
      //

      // first adjust the needed space value
      switch( colinfo[ii].col_type )
      {
          case 2:   /* numeric */

            colinfo[ii].buf_len = 50;
            break;


          case 12:  /* date    */

             colinfo[ii].buf_len = 50;
             break;
      }

      colinfo[ii].pBuf = (char*)malloc( colinfo[ii].buf_len + 1 );
      if ( colinfo[ii].pBuf == NULL ) 
      {
         printf( "panic, no memory left\n" );
         exit(1);
      }

      // nul term that buffer as well
      *(colinfo[ii].pBuf + colinfo[ii].buf_len) = 0;

      //
      // bind the column data to a variable  
      //
      ora_rc = OCIDefineByPos( pOraStmt,
                               &(colinfo[ii].pOraDefine),
                               pOraError,
                               ii+1,                      // position 
                               colinfo[ii].pBuf,          // data buffer
                               colinfo[ii].buf_len,       // data buffer capacity
                               SQLT_CHR,
                               &(colinfo[ii].nOraIndicator),
                               NULL,
                               NULL,
                               OCI_DEFAULT         );

      ORA_CHECK_RC(pOraError,ora_rc)

      //
      // find out the total "rec" len
      // VAR PART
      //
      nRecLen += colinfo[ii].buf_len + ( strlen( colinfo[ii].col_name ) ) + (6+2);

      // allocate that memory for the work buffer
      colinfo[ii].pWork = (char*) malloc( nRecLen + 1 );
      if ( colinfo[ii].pWork == NULL ) 
      {
         printf( "panic, no memory left\n" );
         exit(1);
      }


      /*            */  
      /* debug info */
      /*            */
/*      printf( "size = %u\n", nSize ); */

      printf( "datatype=%d, colname=%s, datasize=%u \n",
              colinfo[ii].col_type,
              colinfo[ii].col_name,
              colinfo[ii].buf_len );      


   	printf( "loop OCIAttrGet: %d \n", ii );
        ORA_CHECK_RC(pOraError,ora_rc)


   } // end for( ii=0...



   //
   // find out the total "rec" len       00001813VNAMmarkus
   // FIX PART                                + +   +
   //                                         | |   |
   //                                    len   Of col data
   //                                    total data
   //                                          starts

   nRecLen = ( nCols * nRecLen ) + 1;

   //
   // alloc memory for the out buffer
   //
   pRec = (char*) malloc( nRecLen );
   if ( pRec == NULL )
   {
      printf( "panic, no memory left\n" );
      exit(1);
   }
   
   
   /*                          */
   /* disp the result          */
   /*                          */

   for( ii=0; ; ii++ )
   {
      // init the out buffer 
      *pRec = 0;

      ora_rc = OCIStmtFetch( pOraStmt, 
                             pOraError,
                             1,                   // how many rows
                             OCI_FETCH_NEXT,
                             OCI_DEFAULT      );

      // check break condition
      if ( ora_rc == OCI_NO_DATA )
      {
         // all rows are processed
         ORA_CHECK_RC(pOraError,ora_rc)
         // printf( "\nOCI_NO_DATA\n" );

         // send no more data
         break;
      }

      if ( ora_rc == OCI_SUCCESS            ||
           ora_rc == OCI_SUCCESS_WITH_INFO     )
      {
         //
         // process each column in the result set
         //
   	for( jj=0; jj < nCols; jj++ )
   	{
            int nTotal = 0;
            int nOffset= 0;

            // remove all blanks on the right side
            //
            rtrim((char*)colinfo[jj].pBuf);

            nOffset = (6+2) + strlen( colinfo[jj].col_name );

            // must be done dynamicly on each field on each row
            nTotal  = nOffset + strlen( colinfo[jj].pBuf );
        
            sprintf( colinfo[jj].pWork, "%06d%02d%s%s",
                     nTotal,
                     nOffset,
                     colinfo[jj].col_name,
                     colinfo[jj].pBuf      );

            //
            // concatenate the columns
            //
            strcat( pRec, colinfo[jj].pWork );
                     
         } // end process each column

         //
         // process the hole "record"
         //
         printf( "NEW RECORD:\n");
         printf( "%s\n", pRec );
        

      }
      else
      {
         // error occured
         printf( "error while processing the result set\n" );
         ORA_CHECK_RC(pOraError,ora_rc)
         exit(1);
      }      
   
   } // end for( ii=0; ;...


   printf( "\nfetched rows: %d\n", ii );


   /*            */			   
   /* do cleanup */  
   /*            */			   

   free( pRec );

   for( ii=0; ii < nCols; ii++ )
   {
      free( colinfo[ii].pBuf );
      free( colinfo[ii].pWork );
      free( colinfo[ii].col_name );
   }


CLEANUP:


   ora_rc = OCISessionEnd( pOraService,
                           pOraError,
                           pOraSession,
                           OCI_DEFAULT    );

   ora_rc = OCIServerDetach( pOraServer, pOraError, OCI_DEFAULT );


   ora_rc = OCIHandleFree( pOraStmt   , (ub4) OCI_HTYPE_STMT    );

   ora_rc = OCIHandleFree( pOraSession, (ub4) OCI_HTYPE_SESSION );
   ora_rc = OCIHandleFree( pOraService, (ub4) OCI_HTYPE_SVCCTX  );
   ora_rc = OCIHandleFree( pOraServer , (ub4) OCI_HTYPE_SERVER  );
   ora_rc = OCIHandleFree( pOraError  , (ub4) OCI_HTYPE_ERROR   );   

   printf( "### end ###\n" );
   return(0);
} 
